#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

void error(const char *msg) {
  perror(msg);
  exit(0);
}

#define BUFLEN 16384
int processCommand(char *buffer, int sockfd);

int main(int argc, char *argv[]) {
  int sockfd, portno, i_arg;
  struct sockaddr_in serv_addr;
  struct hostent *server;
  short persist;

  char buffer[BUFLEN];
  if (argc < 3) {
    fprintf(stderr, "usage %s hostname port [<command> [<command> ...]]\n", argv[0]);
    exit(0);
  }
  portno = atoi(argv[2]);
  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0)
    error("ERROR opening socket");
  server = gethostbyname(argv[1]);
  if (server == NULL) {
    fprintf(stderr, "ERROR, no such host\n");
    exit(0);
  }
  bzero((char *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  bcopy((char *)server->h_addr, (char *)&serv_addr.sin_addr.s_addr, server->h_length);
  serv_addr.sin_port = htons(portno);
  if (connect(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
    error("ERROR connecting");
  i_arg = 3;
  persist = 1;
  for (i_arg = 3; persist && i_arg < argc; i_arg++) {
    sprintf(buffer, "%s\n", argv[i_arg]);
    if (strcmp(buffer, "disconnect\n") == 0)
      persist = 0;
    printf("Processing command: %s", buffer);
    fflush(stdout);
    processCommand(buffer, sockfd);
  }
  while (persist) {
    printf("Please enter the message: ");
    fflush(stdout);
    bzero(buffer, BUFLEN);
    fgets(buffer, BUFLEN - 1, stdin);
    if (strcmp(buffer, "disconnect\n") == 0)
      persist = 0;
    processCommand(buffer, sockfd);
  }
  close(sockfd);
  return 0;
}

int processCommand(char *buffer, int sockfd) {
  int n;
  n = write(sockfd, buffer, strlen(buffer));
  if (n < 0)
    error("ERROR writing to socket");
  bzero(buffer, BUFLEN);
  n = read(sockfd, buffer, BUFLEN - 1);
  if (n < 0)
    error("ERROR reading from socket");
  printf("%s\n", buffer);
  return n;
}
